# Generated by Django 4.0 on 2023-10-13 20:19 by Hesham

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_basin3d', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attributemapping',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='datasource',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



"""django_basin3d URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.10/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
import logging
from basin3d.core.schema.enum import FeatureTypeEnum
from django.urls import re_path
from django.urls import include
from rest_framework import routers

from django_basin3d.models import DataSource
from django_basin3d.synthesis.viewsets import MeasurementTimeseriesTVPObservationViewSet, MonitoringFeatureViewSet
from django_basin3d.views import broker_api_root, monitoring_features_lists
from django_basin3d.viewsets import AttributeMappingViewSet, DataSourceViewSet, ObservedPropertyViewSet

logger = logging.getLogger(__name__)


def get_monitoring_feature_urls():
    """

    :return: list of url objects
    """
    urls = []
    try:
        # iterate over the Datasources and register ViewSets to the router
        # for those models that are defined.
        supported_feature_types = FeatureTypeEnum.values()
        for datasource in DataSource.objects.all():
            viewset_models = []
            plugin = datasource.get_plugin()  # Get the plugin model

            plugin_views = plugin.get_plugin_access()
            for model_name in plugin_views.keys():
                viewset_models.append(model_name.__name__)

            datasource_feature_types = plugin.get_feature_types()
            for feature_type in datasource_feature_types:
                if feature_type in supported_feature_types:
                    ft = ''.join(feature_type.lower().split('_'))
                    path_route = '^monitoringfeature/{}s'.format(ft)
                    urls.extend([
                        re_path(r'{}/$'.format(path_route),
                                MonitoringFeatureViewSet.as_view({'get': 'list'}),
                                name='monitoringfeature-list'),
                        re_path(r'{}\.(?P<format>[a-z0-9]+)/?$'.format(path_route),
                                MonitoringFeatureViewSet.as_view({'get': 'list'}),
                                name='monitoringfeature-list'),
                        re_path(r'{}/(?P<pk>[^/.]+)/$'.format(path_route),
                                MonitoringFeatureViewSet.as_view({'get': '{}s'.format(ft)}),
                                name='monitoringfeature-{}s-detail'.format(ft)),
                        re_path(r'{}/(?P<pk>[^/.]+).(?P<format>[a-z0-9]+)/?'.format(path_route),
                                MonitoringFeatureViewSet.as_view({'get': '{}s'.format(ft)}),
                                name='monitoringfeature-{}s-detail'.format(ft))
                    ])
                    supported_feature_types.remove(feature_type)

    except Exception as e:
        logger.error(e)
    return urls


# Wire up our API using automatic URL routing.
router = routers.DefaultRouter()
router.register(r'datasource', DataSourceViewSet, basename='datasource')
router.register(r'attributemapping', AttributeMappingViewSet, basename='attributemapping')
router.register(r'observedproperty', ObservedPropertyViewSet, basename='observedproperty')

# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    # Broker Root Path View
    re_path(r'^$', broker_api_root, name='broker-api-root'),

    # Views to Dynamic Views
    re_path(r'^measurement_tvp_timeseries/$', MeasurementTimeseriesTVPObservationViewSet.as_view({"get": "list"}),
            name='monitoring-features-list'),
    re_path(r'^monitoringfeature/$', monitoring_features_lists, name='monitoring-features-list'),
]

urlpatterns.extend(get_monitoring_feature_urls())
urlpatterns.append(re_path(r'^', include(router.urls)))



"""
`django_basin3d.models`
***********************

.. currentmodule:: django_basin3d.models

:synopsis: The BASIN-3D Models
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
from __future__ import unicode_literals

from importlib import import_module

from django.db import models


class StringListField(models.TextField):
    """
    StringListField stored delimited strings in the database.

    :param: delimiter
    :type: str
    """

    def __init__(self, *args, **kwargs):
        self.delimiter = ","
        if "delimiter" in kwargs.keys():
            self.delimiter = kwargs["delimiter"]

        super(StringListField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if not value:
            value = []

        if isinstance(value, list) or isinstance(value, tuple):
            return value
        elif isinstance(value, str):
            return value.split(self.delimiter)

        raise ValueError("ListField must be delimited string")

    def get_prep_value(self, value):
        if value is None:
            return value
        else:
            return value

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_db_prep_value(value, None)


class DataSource(models.Model):
    """
    Data Source definition

    Attributes:
        - *id:* string (inherited)
        - *name:* string
        - *id_prefix:* string, prefix that is added to all data source ids
        - *plugin_module:*
        - *plugin_class:*
        - *credentials:*
        - *enabled:*

    """
    name = models.CharField(max_length=20, unique=True, blank=False)
    id_prefix = models.CharField(max_length=5, unique=True, blank=False)
    location = models.TextField(blank=True)
    plugin_module = models.TextField(blank=True)
    plugin_class = models.TextField(blank=True)

    class Meta:
        ordering = ['id_prefix']

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return '<DataSource %r>' % (self.name)

    def get_plugin(self):
        """
        Return the plugin class
        """

        module = import_module(self.plugin_module)
        plugin_class = getattr(module, self.plugin_class)
        from django_basin3d.catalog import CatalogDjango
        return plugin_class(CatalogDjango())


class ObservedProperty(models.Model):
    """
    Defining the properties being observed (measured).
    See https://github.com/BASIN-3D/basin3d/blob/main/basin3d/data/basin3d_observed_property_vocabulary.csv

    Fields:
        - *basin3d_vocab:* string, BASIN-3D observed property vocabulary
        - *full_name:* string, Description of observed property
        - *categories:* List of strings, categories (in order of priority), .
        - *units:* string, units of the observed property
    """

    basin3d_vocab = models.CharField(max_length=50, unique=True, blank=False, primary_key=True)
    full_name = models.CharField(max_length=255)
    categories = StringListField(blank=True, null=True)
    units = models.CharField(max_length=50, blank=False)

    class Meta:
        ordering = ('basin3d_vocab',)

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.description

    def __repr__(self):
        return '<ObservedProperty %r>' % self.basin3d_vocab


class AttributeMapping(models.Model):
    """
    A data class for attribute mappings between datasource vocabularies and BASIN-3D vocabularies.
    These are the associations defined in the datasource (i.e., plugin) mapping file.

    Fields:
         - *attr_type:* Attribute Type; e.g., STATISTIC, RESULT_QUALITY, OBSERVED_PROPERTY; separate compound mappings with ':'
         - *basin3d_vocab:* The BASIN-3D vocabulary; separate compound mappings with ':'
         - *basin3d_desc:* The BASIN-3D vocabulary descriptions; objects or enum
         - *datasource_vocab:* The datasource vocabulary
         - *datasource_desc:* The datasource vocabulary description
         - *datasource:* The datasource of the mapping
    """

    attr_type = models.CharField(max_length=50)
    basin3d_vocab = models.CharField(max_length=50)
    basin3d_desc = models.JSONField()
    datasource_vocab = models.CharField(max_length=50, blank=False)
    datasource_desc = models.TextField(blank=True, null=True)
    datasource = models.ForeignKey(DataSource, on_delete=models.DO_NOTHING)

    class Meta:
        unique_together = ('datasource', 'attr_type', 'datasource_vocab')
        ordering = ('datasource', 'attr_type', 'basin3d_vocab')

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.datasource_vocab



from django_basin3d.models import AttributeMapping, DataSource, ObservedProperty
from django.contrib import admin
from django.contrib.admin import ModelAdmin


@admin.register(DataSource)
class DataSourceAdmin(ModelAdmin):
    list_display = ('name', 'plugin_module', 'plugin_class', 'location')
    fields = ('name', 'location', 'plugin_module', 'plugin_class', 'id_prefix',)
    readonly_fields = ('name', 'id_prefix', 'plugin_module', 'plugin_class')
    actions = None


@admin.register(ObservedProperty)
class ObservedPropertyAdmin(ModelAdmin):
    list_display = ('basin3d_vocab', 'full_name', 'categories', 'units')

    actions = None


@admin.register(AttributeMapping)
class AttributeMappingAdmin(ModelAdmin):
    list_display = ('attr_type', 'basin3d_vocab', 'basin3d_desc', 'datasource_vocab', 'datasource_desc', 'datasource')



"""
`django_basin3d.synthesis.viewsets`
***********************************

.. currentmodule:: django_basin3d.synthesis.viewsets

:synopsis: BASIN-3D Synthesis Model Viewsets (View Controllers) that support the REST API
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

"""
import logging
import pydantic
import typing
from basin3d.core.schema.enum import FeatureTypeEnum
from basin3d.core.schema.query import QueryMeasurementTimeseriesTVP, QueryMonitoringFeature
from rest_framework import status, versioning
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from basin3d.core.models import MeasurementTimeseriesTVPObservation, MonitoringFeature
from basin3d.core.synthesis import DataSourceModelAccess, MeasurementTimeseriesTVPObservationAccess, \
    MonitoringFeatureAccess

from django_basin3d.models import DataSource
from django_basin3d.synthesis.serializers import MeasurementTimeseriesTVPObservationSerializer, \
    MonitoringFeatureSerializer

logger = logging.getLogger(__name__)


def _get_request_feature_type(request):
    """
    Return the feature type if exists in the request
    :param request: request
    otherwise return the text version
    :return: the feature_type in the format specified, None if none exists
    """
    for feature_type in FeatureTypeEnum.values():
        urlpath = request.path_info
        url_feature_type = ''.join(feature_type.lower().split('_'))
        if f'{url_feature_type}s' in urlpath.split('/'):
            return feature_type
    return None


def _convert_str_params_to_list(params: dict, query_class) -> dict:
    """

    :param params:
    :param query_class:
    :return:
    """
    query_arg_types = typing.get_type_hints(query_class)

    for arg, arg_type in query_arg_types.items():
        # arg_type_str = str(arg_type)
        if 'List' not in str(arg_type) or arg not in params.keys():
            continue
        params[arg] = params[arg].split(',')

    return params


class DataSourcePluginViewSet(ViewSet, DataSourceModelAccess):
    """
    Base ViewsSet for all DataSource plugins.  This class extends the
    `Django Rest Framework <https://www.django-rest-framework.org/>`_
    class :class:`rest_framework.viewsets.ViewSet`. These are based on `Django generic views
    <https://docs.djangoproject.com/en/2.2/topics/class-based-views/generic-display/>`_.

    """
    versioning_class = versioning.NamespaceVersioning

    def __init__(self):
        # Override super class
        from django_basin3d.catalog import CatalogDjango
        self._catalog = CatalogDjango()

    @property
    def plugins(self):
        plugins = {}
        for d in DataSource.objects.all():
            plugins[d.id_prefix] = d.get_plugin()
        return plugins

    @property
    def catalog(self):
        return self._catalog

    def list(self, **kwargs) -> Response:
        """
        Return the synthesized plugin results

        :param request: The incoming request object
        :type request: :class:`rest_framework.request.Request`
        :param format: The format to present the data (default is json)
        :return: The HTTP Response
        :rtype: :class:`rest_framework.request.Response`
        """
        items = []

        request = kwargs['request']
        query = kwargs['query']

        itr = super(DataSourcePluginViewSet, self).list(query=query)
        for i in itr:
            items.append(i)

        serializer = self.__class__.serializer_class(items, many=True, context={'request': request})
        synthesis_response = itr.synthesis_response.dict(exclude_unset=True)
        synthesis_response['data'] = serializer.data
        return Response(synthesis_response)

    def retrieve(self, **kwargs) -> Response:
        """
        Retrieve a single object

        :param pk: The primary key
        :return: The HTTP Response
        :rtype: :class:`rest_framework.request.Response`
        """

        request = kwargs['request']
        query = kwargs['query']
        pk = kwargs['pk']

        try:
            item_synthesis_response = super(DataSourcePluginViewSet, self).retrieve(query)

            if not item_synthesis_response or not item_synthesis_response.data:
                return Response({"success": False, "detail": f"There is no detail for {pk}"},
                                status=status.HTTP_404_NOT_FOUND)
            else:

                try:
                    serializer = self.__class__.serializer_class(item_synthesis_response.data, context={'request': request})
                    synthesis_response = item_synthesis_response.dict(exclude_unset=True)
                    synthesis_response['data'] = serializer.data
                    return Response(synthesis_response)
                except Exception as e:
                    logger.error("Plugin error: {}".format(e))

        except Exception as e:
            return Response({'success': False, 'detail': str(e)},
                            status=status.HTTP_404_NOT_FOUND, )


class MonitoringFeatureViewSet(DataSourcePluginViewSet, MonitoringFeatureAccess):
    """
    MonitoringFeature: A feature upon which monitoring is made. OGC Timeseries Profile OM_MonitoringFeature.

    **Synthesis Response**
    This endpoint returns the following synthesis response object.

    ```json
    { "query": {}, "data": [] }
    ```

    **Data Attributes**
    Attribute for each data element from the synthesis response is as follows:

    * *id:* string, Unique feature identifier
    * *name:* string, Feature name
    * *description:* string, Description of the feature
    * *feature_type:* sting, FeatureType: REGION, SUBREGION, BASIN, SUBBASIN, WATERSHED, SUBWATERSHED, SITE, PLOT, HORIZONTAL PATH, VERTICAL PATH, POINT
    * *observed_properties:* list of observed variables made at the feature. Observed property variables are configured via the plugins.
    * *related_sampling_feature_complex:* list of related_sampling features. PARENT features are currently supported.
    * *shape:* string, Shape of the feature: POINT, CURVE, SURFACE, SOLID
    * *coordinates:* location of feature in absolute and/or representative datum
    * *description_reference:* string, additional information about the Feature
    * *related_party:* (optional) list of people or organizations responsible for the Feature
    * *utc_offset:* float, Coordinate Universal Time offset in hours (offset in hours), e.g., +9
    * *url:* url, URL with details for the feature

    **Filter** by the following attributes (/?attribute=parameter&attribute=parameter&...)

    * *datasource (optional):* a single data source id prefix (e.g ?datasource=`datasource.id_prefix`)
    * *parent_feature (optional):* a monitoring feature name

    **Restrict fields**  with query parameter `fields`. (e.g. `?fields=id,name`)
    """
    serializer_class = MonitoringFeatureSerializer
    synthesis_model = MonitoringFeature

    @typing.no_type_check
    def list(self, request: Request, format: str = None) -> Response:
        if not request:
            raise Response({"success": False, "detail": "Request is missing"}, status=status.HTTP_400_BAD_REQUEST)

        feature_type = _get_request_feature_type(request)

        params = request.query_params.dict()
        params = _convert_str_params_to_list(params, QueryMonitoringFeature)

        return super().list(request=request, format=format, query=QueryMonitoringFeature(feature_type=feature_type,
                                                                                         **params))

    @typing.no_type_check
    def retrieve(self, request: Request, pk: str) -> Response:
        if not request:
            return Response({'success': False, 'detail': "Request is missing"},
                            status=status.HTTP_400_BAD_REQUEST, )

        feature_type = _get_request_feature_type(request)

        # retrieve method order: MonitoringFeatureViewSet, DataSourceViewSet, MonitoringFeatureAccess
        # call super force this order based on inherited class ordering
        return super().retrieve(request=request, query=QueryMonitoringFeature(id=pk, feature_type=feature_type), pk=pk)

    @action(detail=True, url_name='regions-detail')
    def regions(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='subregions-detail')
    def subregions(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='basins-detail')
    def basins(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='subbasins-detail')
    def subbasins(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='watersheds-detail')
    def watersheds(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='subwatersheds-detail')
    def subwatersheds(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='sites-detail')
    def sites(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='plots-detail')
    def plots(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='horizontalpaths-detail')
    def horizontalpaths(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='verticalpaths-detail')
    def verticalpaths(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)

    @action(detail=True, url_name='points-detail')
    def points(self, request, pk=None):
        return self.retrieve(request=request, pk=pk)


class MeasurementTimeseriesTVPObservationViewSet(DataSourcePluginViewSet, MeasurementTimeseriesTVPObservationAccess):
    """
    MeasurementTimeseriesTVPObservation: Series of measurement (numerical) observations in
    TVP (time value pair) format grouped by time (i.e., a timeseries).

    **Synthesis Response**
    This endpoint returns the following synthesis response object.

    ```json
    { "query": {}, "data": [] }
    ```

    **Data Attributes**
    Attribute for each data element from the synthesis response is as follows:

    * *id:* string, Observation identifier (optional)
    * *type:* enum, MEASUREMENT_TVP_TIMESERIES
    * *observed_property:* str, BASIN-3D vocabulary for the observation's observed property
    * *datasource:* URL, url of the datasource
    * *sampling_medium:* enum, sampling medium of the observed property (SOLID_PHASE, WATER, GAS, OTHER)
    * *phenomenon_time:* datetime, datetime of the observation, for a timeseries the start and end times can be provided
    * *utc_offset:* float, Coordinate Universal Time offset in hours (offset in hours), e.g., +9
    * *feature_of_interest:* MonitoringFeature obj, feature on which the observation is being made
    * *feature_of_interest_type:* enum (FeatureTypes), feature type of the feature of interest
    * *result:* dict of corresponding lists of TimeValuePairs, the observed values of the observed property being assessed, and (opt) their result_quality,
    * *time_reference_position:* enum, position of timestamp in aggregated_duration (START, MIDDLE, END)
    * *aggregation_duration:* enum, time period represented by observation (YEAR, MONTH, DAY, HOUR, MINUTE, SECOND)
    * *unit_of_measurement:* string, units in which the observation is reported
    * *statistic:* enum, statistical property of the observation result (MEAN, MIN, MAX, TOTAL)
    * *result_quality:* list of enum, quality assessment of the result enum (VALIDATED, UNVALIDATED, SUSPECTED, REJECTED, ESTIMATED)

    **Filter** by the following attributes (?attribute=parameter,parameter&attribute=parameter&...):

    * *monitoring_feature (required):* comma separated list of monitoring_features ids
    * *observed_property (required):* comma separated list of observed property basin3d vocabularies
    * *start_date (required):* date YYYY-MM-DD
    * *end_date (optional):* date YYYY-MM-DD
    * *aggregation_duration (default: DAY):* enum (YEAR|MONTH|DAY|HOUR|MINUTE|SECOND|NONE)
    * *statistic (optional):* comma separated list of statistic enum(s) (MEAN|MIN|MAX|INSTANTANEOUS)
    * *result_quality (optional):* comma separated list of result quality enum(s) enum (VALIDATED|UNVALIDATED|SUSPECTED|REJECTED|ESTIMATED)
    * *sampling_medium (optional):* comma separated list of sampling medium enum(s) (SOLID_PHASE|WATER|GAS|OTHER)
    * *datasource (optional):* a single data source id prefix (e.g ?datasource=`datasource.id_prefix`)

    **Restrict fields** with query parameter `fields`. (e.g. `?fields=id,name`)

    """
    serializer_class = MeasurementTimeseriesTVPObservationSerializer
    synthesis_model = MeasurementTimeseriesTVPObservation

    @action(detail=False, url_path='measurement_tvp_timeseries', url_name='measurementtvptimeseries-list', methods=['GET'])
    def list(self, request: Request, format: str = None) -> Response:
        if not request:
            return Response({'success': False, 'detail': "Request is missing"},
                            status=status.HTTP_400_BAD_REQUEST, )

        params = request.query_params.dict()
        params = _convert_str_params_to_list(params, QueryMeasurementTimeseriesTVP)

        try:
            return super().list(request=request, format=format, query=QueryMeasurementTimeseriesTVP(**params))
        except pydantic.ValidationError as exec_info:
            return Response({'success': False, 'detail': "Missing or invalid search criteria",
                             "errors": exec_info.errors()},
                            status=status.HTTP_400_BAD_REQUEST, )



"""
`django_basin3d.synthesis`
**************************

.. currentmodule:: django_basin3d.synthesis

:synopsis: The high level BASIN-3D synthesis models.
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>
:module author: Charuleka Varadharajan <cvaradharajan@lbl.gov>


* :py:mod:`~django_basin3d.synthesis.serializers` - Serializers that render :py:mod:`basin3d.core.models` from Python objects to `JSON` and back again.
* :py:mod:`~django_basin3d.synthesis.viewsets` - Controllers for BASIN-3D REST api

"""
from . import serializers

__all__ = ['serializers']



"""

.. currentmodule:: django_basin3d.serializers

:platform: Unix, Mac
:synopsis: BASIN-3D Serializers
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

About Django Serializers:

    Serializers allow complex data such as querysets and model instances to be converted
    to native Python datatypes that can then be easily rendered into JSON, XML or other
    content types. Serializers also provide deserialization, allowing parsed data to be
    converted back into complex types, after first validating the incoming data.
    https://www.django-rest-framework.org/api-guide/serializers/
"""
from django_basin3d.models import DataSource, ObservedProperty, AttributeMapping
from rest_framework.reverse import reverse
from rest_framework import serializers


class ChooseFieldsSerializerMixin(object):
    """
    A serializer that dynamically sets fields
    """

    def __init__(self, *args, **kwargs):

        # Instantiate the serializer superclass
        super(ChooseFieldsSerializerMixin, self).__init__(*args, **kwargs)

        if 'request' in self.context:  # type: ignore
            self.handle_fields(self.context['request'])  # type: ignore

    def handle_fields(self, request=None):
        """
        Restrict the fields by those in the request
        :param request:
        :return:
        """
        if request:
            fields = request.query_params.get('fields')
            if fields and len(fields) > 0:
                field_set = set(fields.split(","))

                # Remove the fields not in the intersection
                for field in set(self.fields.keys()).difference(field_set):  # type: ignore
                    self.fields.pop(field)  # type: ignore


class DelimitedListField(serializers.ListField):
    """
    Convert a delimited string field to a list
    """

    child = serializers.CharField()

    def __init__(self, *args, delimiter=",", **kwargs):
        super(DelimitedListField, self).__init__(*args, **kwargs)
        self.delimiter = delimiter

    def to_representation(self, data):
        """
        List of object instances -> List of dicts of primitive datatypes.
        """
        data_values = data.split(self.delimiter)
        return [self.child.to_representation(item) if item is not None else None for item in
                data_values]


class DataSourceSerializer(serializers.HyperlinkedModelSerializer):
    """
    Data Source serializer that converts a models.DataSource
    """

    url = serializers.SerializerMethodField()
    attribute_mapping = serializers.SerializerMethodField()
    observed_property = serializers.SerializerMethodField()
    check = serializers.SerializerMethodField()

    def get_url(self, obj):
        """

        :param obj:
        :return:
        """
        url_kwargs = {'id_prefix': obj.id_prefix, }
        return "{}".format(reverse('datasource-detail', kwargs=url_kwargs, request=self.context["request"], ))

    def get_attribute_mapping(self, obj):
        """
        Return the url for the attribute mapping associated with the current datasource
        :param obj:
        :return:
        """
        format = None
        if "format" in self.context["request"].query_params:
            format = self.context["request"].query_params["format"]

        url_kwargs = {"id_prefix": obj.id_prefix, }

        return reverse("{}-attribute-mapping".format(obj.__class__.__name__.lower()), kwargs=url_kwargs,
                       request=self.context["request"], format=format)

    def get_observed_property(self, obj):
        """
        Return the url for the observed property associated with the current datasource
        :param obj:
        :return:
        """
        format = None
        if "format" in self.context["request"].query_params:
            format = self.context["request"].query_params["format"]

        url_kwargs = {"id_prefix": obj.id_prefix, }

        return reverse("{}-observed-property".format(obj.__class__.__name__.lower()), kwargs=url_kwargs,
                       request=self.context["request"], format=format)

    def get_check(self, obj):
        """
        Check the data source
        :param obj:
        :return:
        """
        url_kwargs = {'id_prefix': obj.id_prefix, }
        return "{}check/".format(reverse('datasource-detail', kwargs=url_kwargs, request=self.context["request"], ))

    class Meta:
        model = DataSource
        depth = 1
        fields = ('url', 'name', 'location', 'id_prefix',
                  'attribute_mapping', 'observed_property', 'check')
        read_only_fields = ('name', 'location', 'id_prefix',
                            'attribute_mapping', 'observed_property', 'check')
        lookup_field = 'name'


class ObservedPropertySerializer(serializers.HyperlinkedModelSerializer):
    """
    Observed Property Serializer
    """

    categories = DelimitedListField()

    def get_categories(self, obj):
        return obj.categories

    class Meta:
        model = ObservedProperty
        depth = 2
        fields = ('url', 'basin3d_vocab', 'full_name', 'categories', 'units')


class AttributeMappingSerializer(serializers.HyperlinkedModelSerializer):
    """
    Attribute Mapping Serializer
    """

    datasource = serializers.SerializerMethodField()
    basin3d_desc = serializers.SerializerMethodField()

    def get_datasource(self, obj):
        """
        Return the url for the data sources associated with the current variable
        :param obj:
        :return:
        """
        url_kwargs = {'id_prefix': obj.datasource.id_prefix, }
        return "{}".format(reverse('datasource-detail', kwargs=url_kwargs, request=self.context["request"], ))

    def get_basin3d_desc(self, obj):
        """
        Return a list of str for enum and url for observed properties
        :param obj:
        :return:
        """
        return obj.basin3d_desc

    class Meta:
        model = AttributeMapping
        depth = 2
        fields = ('url', 'attr_type', 'basin3d_vocab', 'basin3d_desc', 'datasource_vocab', 'datasource_desc', 'datasource')



"""
`django_basin3d.catalog`
************************

.. currentmodule:: django_basin3d.catalog

:synopsis: The Django BASIN-3D Catalog
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
import importlib
import json
import logging
from typing import Iterator, List, Optional, Union

from django.conf import settings
from django.db import IntegrityError, OperationalError
from django.db.models import Q

from basin3d.core.catalog import CatalogBase, CatalogException
from basin3d.core.models import ObservedProperty, AttributeMapping
from basin3d.core.plugin import PluginMount
from basin3d.core.schema.enum import BaseEnum, MappedAttributeEnum, MAPPING_DELIMITER, NO_MAPPING_TEXT, set_mapped_attribute_enum_type

logger = logging.getLogger(__name__)


class CatalogDjango(CatalogBase):

    def __init__(self, variable_filename: str = 'basin3d_observed_property_vocabulary.csv'):
        super().__init__(variable_filename)

    def is_initialized(self) -> bool:
        """Has the catalog been initialized?"""

        try:
            from django_basin3d.models import DataSource
            datasources = DataSource.objects.count()
            if isinstance(datasources, int):
                return datasources > 0
            logger.debug('Catalog not initialized')
            return False
        except ImportError:
            return False

    def _convert_django_observed_property(self, django_opv) -> Optional[ObservedProperty]:
        """
        Convert django observed property variable to basin3d
        :param django_opv:
        :return:
        """
        if django_opv:
            return ObservedProperty(
                basin3d_vocab=django_opv.basin3d_vocab,
                full_name=django_opv.full_name,
                categories=django_opv.categories.split(","),
                units=django_opv.units
            )
        return None

    def _convert_django_attribute_mapping(self, django_am) -> Optional[AttributeMapping]:
        """
        Convert django attribute_mapping
        :param django_am:
        :return:
        """
        if django_am:

            attr_type_list = django_am.attr_type.split(MAPPING_DELIMITER)

            if isinstance(django_am.basin3d_desc, list):
                basin3d_desc_list = django_am.basin3d_desc
            else:
                # It should always be a list
                try:
                    basin3d_desc_list = json.loads(django_am.basin3d_desc)
                except Exception as e:
                    raise e

            basin3d_desc = []

            for attr_type, desc in zip(attr_type_list, basin3d_desc_list):
                if attr_type == MappedAttributeEnum.OBSERVED_PROPERTY.value:
                    op = ObservedProperty(
                        basin3d_vocab=desc.get('basin3d_vocab'),
                        full_name=desc.get('full_name'),
                        categories=desc.get('categories'),
                        units=desc.get('units')
                    )
                    basin3d_desc.append(op)
                elif attr_type in MappedAttributeEnum.values():
                    attr_enum_class = set_mapped_attribute_enum_type(attr_type)
                    attr_type_enum = getattr(attr_enum_class, desc)
                    basin3d_desc.append(attr_type_enum)
                else:
                    basin3d_desc.append(desc)

            return AttributeMapping(
                attr_type=django_am.attr_type,
                basin3d_vocab=django_am.basin3d_vocab,
                basin3d_desc=basin3d_desc,
                datasource_vocab=django_am.datasource_vocab,
                datasource_desc=django_am.datasource_desc,
                datasource=django_am.datasource
            )

        return None

    def _convert_basin3d_attr_mapping_basin3d_desc(self, basin3d_desc: list) -> list:
        json_ready_basin3d_desc = []

        for desc in basin3d_desc:
            if isinstance(desc, ObservedProperty):
                json_ready_basin3d_desc.append(desc.to_dict())
            elif isinstance(desc, BaseEnum):
                json_ready_basin3d_desc.append(desc.value)
            else:
                json_ready_basin3d_desc.append(desc)

        return json_ready_basin3d_desc

    def _get_observed_property(self, basin3d_vocab) -> Optional[ObservedProperty]:
        """
        Access a single observed property variable

        :param basin3d_vocab: the observed property name
        :return:
        """
        from django_basin3d import models as django_models

        try:
            opv = django_models.ObservedProperty.objects.get(basin3d_vocab=basin3d_vocab)
            return self._convert_django_observed_property(opv)
        except django_models.ObservedProperty.DoesNotExist:
            return None
        except Exception as e:
            if not e.__class__.__name__ == 'DoesNotExist':
                raise e
            return None

    def _get_attribute_mapping(self, datasource_id, attr_type, basin3d_vocab, datasource_vocab, **kwargs) -> Optional[AttributeMapping]:
        """

        :param datasource_id:
        :param attr_type:
        :param basin3d_vocab:
        :param datasource_vocab:
        :param kwargs:
        :return:
        """
        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        from django_basin3d import models as django_models

        try:
            opv = django_models.AttributeMapping.objects.get(
                datasource__name=datasource_id, attr_type=attr_type, basin3d_vocab=basin3d_vocab, datasource_vocab=datasource_vocab)
            return self._convert_django_attribute_mapping(opv)
        except django_models.ObservedProperty.DoesNotExist:
            return None
        except Exception as e:
            if not e.__class__.__name__ == 'DoesNotExist':
                raise e
            return None

    def find_observed_property(self, basin3d_vocab) -> Optional[ObservedProperty]:
        """
                Return the :class:`basin3d.models.ObservedProperty` object for the BASIN-3D vocabulary specified.

                :param basin3d_vocab: BASIN-3D vocabulary
                :return: a :class:`basin3d.models.ObservedProperty` object
                """
        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        return self._get_observed_property(basin3d_vocab)

    def find_observed_properties(self, basin3d_vocab: Optional[List[str]] = None) -> Iterator[Optional[ObservedProperty]]:
        """
        Report the observed_properties available based on the BASIN-3D vocabularies specified. If no BASIN-3D vocabularies are specified, then return all observed properties available.

        :param basin3d_vocab: list of the BASIN-3D observed properties
        :return: generator that yields :class:`basin3d.models.ObservedProperty` objects
        """
        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        from django_basin3d import models as django_models

        if not basin3d_vocab:
            for opv in django_models.ObservedProperty.objects.all():
                yield self._convert_django_observed_property(opv)
        else:
            for b3d_vocab in basin3d_vocab:
                opv = self._get_observed_property(b3d_vocab)
                if opv is not None:
                    yield opv

    def find_datasource_attribute_mapping(self, datasource_id: str, attr_type: str, datasource_vocab: str) -> Optional[AttributeMapping]:
        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        # Consider checking args for a value

        from django_basin3d import models as django_models

        # Setup the search parameters
        query_params = {
            'datasource__name': datasource_id,
            'attr_type__contains': attr_type,
            'datasource_vocab': datasource_vocab
        }

        msg = (f'No mapping was found for attr: "{attr_type}" and for datasource vocab: "{datasource_vocab}" '
               f'in datasource: "{datasource_id}".')

        try:
            ds = django_models.DataSource.objects.get(name=datasource_id)
        except django_models.DataSource.DoesNotExist:
            msg = f'No Data Source "{datasource_id}" found.'
            ds = django_models.DataSource(name=None, location=None, id_prefix=None, plugin_module=None, plugin_class=None)
        except Exception as e:
            if e.__class__.__name__ not in ['DoesNotExist']:
                raise e

        # set up empty AttributeMapping in case where mapping is not found or another error occurs
        attr_mapping = AttributeMapping(attr_type=attr_type, basin3d_vocab=NO_MAPPING_TEXT, basin3d_desc=[],
                                        datasource_vocab=datasource_vocab, datasource_desc=msg, datasource=ds)

        try:
            attr_mapping = django_models.AttributeMapping.objects.get(**query_params)
        except django_models.AttributeMapping.DoesNotExist:
            return attr_mapping
        except django_models.AttributeMapping.MultipleObjectsReturned:
            msg = (f'Multiple mappings found for attr: "{attr_type}" and datasource vocab: "{datasource_vocab}" '
                   f'in datasource: "{datasource_id}". This should never happen.')
            attr_mapping.datasource_desc = msg
            return attr_mapping
        except Exception as e:
            if e.__class__.__name__ not in ['DoesNotExist', 'MultipleObjectsReturned']:
                raise e

        return self._convert_django_attribute_mapping(attr_mapping)

    def find_attribute_mappings(self, datasource_id: str = None, attr_type: str = None, attr_vocab: Union[str, List] = None,
                                from_basin3d: bool = False) -> Iterator[AttributeMapping]:

        if not self.is_initialized():
            raise CatalogException("Datasource catalog has not been initialized")

        def construct_attr_vocab_query(attr_vocab_list, is_from_basin3d):
            query = Q(_connector=Q.OR)
            for a_vocab in attr_vocab_list:
                if not is_from_basin3d:
                    query.add(('datasource_vocab__exact', a_vocab), conn_type=Q.OR)
                elif MAPPING_DELIMITER in a_vocab:
                    query.add(('basin3d_vocab__regex', a_vocab), conn_type=Q.OR)
                else:
                    query.add(('basin3d_vocab__exact', a_vocab), conn_type=Q.OR)
                    query.add(('basin3d_vocab__regex', f'.*:{a_vocab}'), conn_type=Q.OR)
                    query.add(('basin3d_vocab__regex', f'{a_vocab}:.*'), conn_type=Q.OR)
                    query.add(('basin3d_vocab__regex', f'.*:{a_vocab}:.*'), conn_type=Q.OR)
            return query

        from django_basin3d import models as django_models

        query_params = []

        if datasource_id is not None:
            try:
                django_models.DataSource.objects.get(name=datasource_id)
            except django_models.DataSource.DoesNotExist:
                logger.warning(f'No datasource for datasource_id {datasource_id} was found. Check plugin initialization')
                yield None
            except Exception as e:
                if e.__class__.__name__ not in ['DoesNotExist']:
                    raise CatalogException(e)

            query_params.append(Q(datasource__name=datasource_id))

        if attr_type is not None:
            if attr_type not in MappedAttributeEnum.values():
                logger.warning(f'Attribute type {attr_type} is invalid')
                yield None

            query_params.append(Q(attr_type__contains=attr_type))

        if attr_vocab:
            if isinstance(attr_vocab, str):
                attr_vocab = [attr_vocab]
            elif not isinstance(attr_vocab, List):
                raise CatalogException("attr_vocab must be a str or list")
            attr_vocab_query = construct_attr_vocab_query(attr_vocab, from_basin3d)
            query_params.append(attr_vocab_query)

        try:
            attr_mappings = django_models.AttributeMapping.objects.filter(*query_params)
        except django_models.AttributeMapping.DoesNotExist:
            vocab_source_type = 'datasource'
            if from_basin3d:
                vocab_source_type = 'BASIN-3D'
            logger.info(f'No mapping was found for attr: "{attr_type}" and for {vocab_source_type} vocab: "{attr_vocab}" '
                        f'in datasource: "{datasource_id}".')
            pass
        except Exception as e:
            if e.__class__.__name__ not in ['DoesNotExist']:
                raise e

        for attr_mapping in attr_mappings:
            yield self._convert_django_attribute_mapping(attr_mapping)

    def _init_catalog(self, **kwargs):
        """
        Initialize the catalog database

        :return:
        """
        if not self.is_initialized():
            from django_basin3d import models as django_models

            # Now create the Datasource objects in the data base
            from basin3d.core.plugin import PluginMount
            for name, plugin in PluginMount.plugins.items():
                module_name = plugin.__module__
                class_name = plugin.__name__

                logger.info("Loading Plugin = {}.{}".format(module_name, class_name))

                try:
                    datasource = django_models.DataSource.objects.get(name=plugin.get_meta().id)
                except django_models.DataSource.DoesNotExist:
                    logger.info("Registering NEW Data Source Plugin '{}.{}'".format(module_name, class_name))
                    datasource = django_models.DataSource()
                    if hasattr(plugin.get_meta(), "connection_class"):
                        datasource.credentials = plugin.get_meta().connection_class.get_credentials_format()

                # Update the datasource
                datasource.name = plugin.get_meta().id
                datasource.location = plugin.get_meta().location
                datasource.id_prefix = plugin.get_meta().id_prefix
                datasource.plugin_module = module_name
                datasource.plugin_class = class_name
                datasource.save()
                logger.info("Updated Data Source '{}'".format(plugin.get_meta().id))

    def _insert(self, record):
        """
        :param record:
        """
        from django_basin3d import models as django_models

        if self.is_initialized():
            if isinstance(record, ObservedProperty):
                try:
                    p = django_models.ObservedProperty()
                    p.basin3d_vocab = record.basin3d_vocab
                    p.full_name = record.full_name
                    p.categories = ",".join(record.categories)  # type: ignore
                    p.units = record.units
                    p.save()
                    logger.info(f'inserted {record.basin3d_vocab}')

                except IntegrityError as ie:
                    # This object has already been loaded
                    logger.debug(f'Integrity error for OP: {ie}')
                    pass

                except Exception as e:
                    logger.warning("Error Registering ObservedProperty '{}': {}".format(record.basin3d_vocab, str(e)))

            elif isinstance(record, AttributeMapping):
                try:
                    ds_name = django_models.DataSource.objects.get(name=record.datasource.id)

                    record_basin3d_desc = self._convert_basin3d_attr_mapping_basin3d_desc(record.basin3d_desc)

                    p = django_models.AttributeMapping()
                    p.datasource = ds_name
                    p.attr_type = record.attr_type
                    p.basin3d_vocab = record.basin3d_vocab
                    p.basin3d_desc = record_basin3d_desc
                    p.datasource_vocab = record.datasource_vocab
                    p.datasource_desc = record.datasource_desc
                    p.save()
                    logger.info(f'inserted {record.datasource_vocab} mapping attribute')

                except IntegrityError:
                    # This object has already been loaded
                    logger.info(f'Warning: skipping AttributeMapping "{record.basin3d_vocab}". Already loaded.')
                    pass

                except Exception as e:
                    logger.info(f'Error Registering AttributeMapping "{record.basin3d_vocab}": {str(e)}')
        else:
            raise CatalogException('Could not insert record. Catalog not initialize')


def load_data_sources(sender, **kwargs):
    """
    Load the Broker data sources from the registered plugins.

    :param sender:
    :param kwargs:
    :return:
    """
    # Load all the plugins found in apps
    for django_app in settings.INSTALLED_APPS:
        try:
            importlib.import_module("{}.plugins".format(django_app))
            logger.info(f"Loaded {django_app} plugins")

            catalog = CatalogDjango()
            catalog.initialize([v(catalog) for v in PluginMount.plugins.values()])
        except ImportError as e:
            logger.error(f'import_error: {e}')
            pass
        except OperationalError as oe:
            logger.error(f"Operational Error '{oe}'' - Most likely happens on a reverse migration.")



"""
`django_basin3d.viewsets`
*************************

.. currentmodule:: django_basin3d.viewsets

:platform: Unix, Mac
:synopsis: BASIN-3D ViewSets
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top


"""
import django_filters
from rest_framework.decorators import action

from basin3d.core.schema.enum import MAPPING_DELIMITER, MappedAttributeEnum
from django_basin3d import get_url
from django_basin3d.models import AttributeMapping, DataSource, ObservedProperty
from django_basin3d.serializers import DataSourceSerializer, ObservedPropertySerializer, AttributeMappingSerializer
from rest_framework import status
from rest_framework import viewsets
from rest_framework.response import Response


class DataSourceViewSet(viewsets.ReadOnlyModelViewSet):
    """
        Returns a list of all Data Sources available to the BASIN-3D service

        **Properties**

        * *url:* url, Endpoint for Data Source
        * *name:* string, Unique name for the Data Source
        * *location:* string, Location of the Data Source
        * *id_prefix:* string, unique id prefix for all Data Source ids
        * *attribute_mapping:* url, List of Attribute Mappings that are mapped for Data Source
        * *observed_property:* url, List of Observed Properties that are mapped for Data Source
        * *check:* url, Validate the Data Source connection

    """
    queryset = DataSource.objects.all()
    serializer_class = DataSourceSerializer
    lookup_field = 'id_prefix'

    @action(detail=True)
    def check(self, request, id_prefix=None):
        """
        Determine if Data Source is available
        :param request:
        :param id_prefix:
        :return:
        """

        datasource = self.get_object()

        plugin = datasource.get_plugin()

        if hasattr(plugin.get_meta(), "connection_class"):
            http_auth = plugin.get_meta().connection_class(datasource)

            try:
                http_auth.login()
                return Response(data={"message": "Login to {} data source was successful".format(datasource.name),
                                      "success": True},
                                status=status.HTTP_200_OK)
            except Exception as e:
                return Response(data={"message": str(e), "success": False},
                                status=status.HTTP_200_OK)

            finally:
                http_auth.logout()
        else:
            try:
                response = get_url("{}".format(datasource.location))

                if response.status_code == status.HTTP_200_OK:
                    return Response(
                        data={"message": "Response from {} data source was successful".format(datasource.name),
                              "success": True},
                        status=status.HTTP_200_OK)
                else:
                    return Response(
                        data={
                            "message": "Response from {} data source returns HTTP status {}".format(datasource.name,
                                                                                                    response.status_code),
                            "success": True},
                        status=status.HTTP_200_OK)

            except Exception as e:
                return Response(data={"message": str(e), "success": False},
                                status=status.HTTP_200_OK)

    @action(detail=True)  # Custom Route for an association
    def attribute_mapping(self, request, id_prefix=None):
        """
        Retrieve the Attribute Mappings for a Data Source.

        Maps to /datasource/{id_prefix}/attributemapping/

        :param request:
        :param id_prefix:
        :return:
        """
        params = AttributeMapping.objects.filter(datasource__id_prefix=id_prefix)

        # `HyperlinkedRelatedField` req:w
        # uires the request in the
        # serializer context. Add `context={'request': request}`
        # when instantiating the serializer.

        # Then just serialize and return it!
        serializer = AttributeMappingSerializer(params, many=True, context={'request': request})
        return Response(serializer.data)

    @action(detail=True)  # Custom Route for an association
    def observed_property(self, request, id_prefix=None):
        """
        Retrieve the Observed Properties mapped for the current Data Source.

        Maps to /datasource/{id_prefix}/observedproperty/

        :param request:
        :param id_prefix:
        :return:
        """
        op_name = MappedAttributeEnum.OBSERVED_PROPERTY.value
        params = AttributeMapping.objects.filter(datasource__id_prefix=id_prefix, attr_type__contains=op_name)
        v = []
        for am in params:
            attr_type_list = am.attr_type.split(MAPPING_DELIMITER)
            vocab_list = am.basin3d_vocab.split(MAPPING_DELIMITER)
            for attr, vocab in zip(attr_type_list, vocab_list):
                if attr == op_name:
                    v.append(ObservedProperty.objects.get(pk=vocab))

        # `HyperlinkedRelatedField` requires the request in the
        # serializer context. Add `context={'request': request}`
        # when instantiating the serializer.

        # Then just serialize and return it!
        serializer = ObservedPropertySerializer(v, many=True, context={'request': request})
        return Response(serializer.data)


class AttributeMappingViewSet(viewsets.ReadOnlyModelViewSet):
    """
        Returns a list of the Attribute Mappings for the registered Data Sources.

        **Properties**

        * *url:* url, endpoint for Attribute Mapping
        * *attr_type:* string, attribute mapping type
        * *basin3d_vocab:* string, observed property vocabulary
        * *basin3d_desc:* list, Observed Property (dict) or enum (str)
        * *datasource_vocab:* string, datasource vocabulary
        * *datasource_desc:* string, datasource description
        * *datasource:* string, data source defining the observed property

    """
    queryset = AttributeMapping.objects.all()
    serializer_class = AttributeMappingSerializer
    filter_backends = (django_filters.rest_framework.DjangoFilterBackend,)


class ObservedPropertyViewSet(viewsets.ReadOnlyModelViewSet):
    """
        Returns a list of available BASIN-3D Observed Properties

        **Properties**

        * *url:* url, Endpoint for the observed property vocabulary
        * *basin3d_vocab:* string, Unique BASIN-3D observed property vocabulary
        * *full_name:* string, Descriptive name
        * *categories:* list of strings, Categories of which the variable is a member, listed in hierarchical order
        * *units:* string, units

    """
    queryset = ObservedProperty.objects.all()
    serializer_class = ObservedPropertySerializer
    filter_backends = (django_filters.rest_framework.DjangoFilterBackend,)



from django.apps import AppConfig
from django.db.models.signals import post_migrate

from django_basin3d.catalog import load_data_sources


class Basin3DConfig(AppConfig):
    name = 'django_basin3d'

    def ready(self):
        # Execute the post migration scripts
        post_migrate.connect(load_data_sources, sender=self)



from django import template

register = template.Library()


@register.filter(name='split')
def split(value, arg):
    return value.rstrip(arg).lstrip(arg).split(arg)


@register.filter(name='breadcrumb_convert')
def breadcrumb_name(value, url):
    values = url.rstrip("/").lstrip("/").split("/")
    if len(values) > 2:
        return values[-1].title()
    else:
        return value



"""
`django_basin3d.synthesis.serializers`
**************************************

.. currentmodule:: django_basin3d.synthesis.serializers

:synopsis: The BASIN-3D Synthesis Model Serializers
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

Serializers that render :py:mod:`basin3d.core.models` from Python objects to `JSON` and back again.

"""
import logging
from basin3d.core.schema.enum import FeatureTypeEnum, NO_MAPPING_TEXT
from numbers import Number
from typing import List

from django.utils.datetime_safe import datetime
from rest_framework import serializers
from rest_framework.reverse import reverse

from django_basin3d.serializers import ChooseFieldsSerializerMixin

logger = logging.getLogger(__name__)


class TimestampField(serializers.DateTimeField):
    """
    Extends :class:`rest_framework.serializers.DateTimeField` to handle
    numeric epoch times.

    """

    def to_representation(self, value):
        """
        If specified value is an epoch time, convert it first.

        :param value:
        :return:
        """

        # Handle epoch time
        timestamp = None
        if isinstance(value, str) and value.isdigit():
            timestamp = int(value)
        elif isinstance(value, Number):
            timestamp = int(str(value))

        # ToDo: add additional time formats
        if timestamp:
            value = datetime.fromtimestamp(timestamp).isoformat()

        return value


class ReadOnlySynthesisModelField(serializers.Field):
    """
    A generic field that can be used against any serializer
    """

    def __init__(self, serializer_class, **kwargs):
        self.serializer_class = serializer_class
        super(ReadOnlySynthesisModelField, self).__init__(read_only=True, **kwargs)

    def to_internal_value(self, data):
        raise NotImplementedError

    def to_representation(self, obj):
        serializer = self.serializer_class(obj, context=self.context)
        return serializer.data


class FloatField(serializers.FloatField):
    """
    A Float field that can handle empty strings
    """

    def to_representation(self, value):
        """to float representation"""
        if not value:
            return None
        return float(value)


class MappedAttributeField(serializers.Field):
    """
    A field to handle the special case of a Mapped Attribute
    """

    def to_internal_value(self, data):
        raise NotImplementedError

    def to_representation(self, obj):
        if not obj:
            return None
        basin3d_vocab = obj.get_basin3d_vocab()
        return basin3d_vocab


class IdUrlSerializerMixin(object):
    """
    Serializer Mixin to support Hypermedia as the Engine of Application State (HATEOAS).
    """

    def __init__(self, *args, **kwargs):
        # Instantiate the serializer superclass
        super(IdUrlSerializerMixin, self).__init__(*args, **kwargs)

        self.fields["url"] = serializers.SerializerMethodField()  # type: ignore

    def get_url(self, obj):
        """
        Get the Site url based on the current context
        :param obj: an object instance
        :return: An URL to the current object instance
        """
        if "request" in self.context and self.context["request"]:  # type: ignore
            return reverse(viewname='{}-detail'.format(obj.__class__.__name__.lower()),
                           kwargs={'pk': obj.id},
                           request=self.context["request"], )  # type: ignore


class PersonSerializer(serializers.Serializer):
    """ Serializes a :class:`basin3d.core.models.Person`"""

    first_name = serializers.CharField(read_only=True)
    last_name = serializers.CharField(read_only=True)
    email = serializers.EmailField(read_only=True)
    institution = serializers.CharField(read_only=True)
    role = serializers.CharField(read_only=True)


class VerticalCoordinateSerializer(serializers.Serializer):
    """ Serializes a :class:`basin3d.core.models.VerticalCoordinate` and its child classes """

    value = serializers.FloatField(read_only=True)
    resolution = serializers.FloatField(read_only=True)
    distance_units = serializers.CharField(read_only=True)
    datum = serializers.CharField(read_only=True)
    type = serializers.CharField(read_only=True)


class HorizonatalCoordinateSerializer(serializers.Serializer):
    """ Serializes a :class:`basin3d.core.models.HorizonatalCoordinate` and its child classes """

    # Base Fields
    x = FloatField(read_only=True)
    y = FloatField(read_only=True)
    datum = serializers.CharField(read_only=True)
    type = serializers.CharField(read_only=True)

    # Geographic Fields
    latitude = FloatField(read_only=True)
    longitude = FloatField(read_only=True)
    units = serializers.CharField(read_only=True)

    # Geographic Fields
    FIELDS_HORIZONTAL = {'X', 'Y'}
    FIELDS_GEOGRAPHIC = {'LATITUDE', 'LONGITUDE'}

    def __init__(self, *args, **kwargs):
        """
        Override `serializers.BaseSerializer.__init__()` to modify the fields outputted. This depends on the
        type of coordinate classes in :module:`basin3d.core.models`

        See the synthesis classes for a list of attributes:
        * :class:`basin3d.core.models.GeographicCoordinate`

        :param args:
        :param kwargs:
        """
        super(HorizonatalCoordinateSerializer, self).__init__(*args, **kwargs)

        field_to_remove = set()
        field_to_remove.update(self.FIELDS_HORIZONTAL)
        field_to_remove.update(self.FIELDS_GEOGRAPHIC)
        instance = None
        if "instance" in kwargs:
            instance = kwargs["instance"]
        elif len(args) >= 1:
            if args[0] and isinstance(args[0], (list, tuple)) and not isinstance(args[0], str):
                instance = args[0][0]
            else:
                instance = args[0]

        if instance:

            from basin3d.core.models import GeographicCoordinate
            if isinstance(instance, GeographicCoordinate):
                field_to_remove -= self.FIELDS_GEOGRAPHIC

        # remove unneeded fields
        for field in field_to_remove:
            if field in self.fields:
                self.fields.pop(field)


class AbsoluteCoordinateSerializer(ChooseFieldsSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.AbsoluteCoordinate`
    """

    horizontal_position = serializers.ListSerializer(
        child=ReadOnlySynthesisModelField(serializer_class=HorizonatalCoordinateSerializer))
    vertical_extent = serializers.ListSerializer(
        child=ReadOnlySynthesisModelField(serializer_class=VerticalCoordinateSerializer))

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)


class RepresentativeCoordinateSerializer(ChooseFieldsSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.RepresentativeCoordinate`
    """

    representative_point = ReadOnlySynthesisModelField(serializer_class=AbsoluteCoordinateSerializer)
    representative_point_type = serializers.CharField(read_only=True)
    vertical_position = ReadOnlySynthesisModelField(serializer_class=VerticalCoordinateSerializer)

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)


class CoordinateSerializer(ChooseFieldsSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.Coordinate`
    """

    absolute = ReadOnlySynthesisModelField(serializer_class=AbsoluteCoordinateSerializer)
    representative = ReadOnlySynthesisModelField(serializer_class=RepresentativeCoordinateSerializer)

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)


class RelatedSamplingFeatureSerializer(ChooseFieldsSerializerMixin, IdUrlSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.RelatedSamplingFeature`
    """
    related_sampling_feature = serializers.CharField(read_only=True)
    related_sampling_feature_type = serializers.CharField(read_only=True)
    role = serializers.CharField(read_only=True)

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)
        super(self.__class__, self).__init__(*args, **kwargs)

    def get_url(self, obj):
        """
        Get the  url based on the current context
        :param obj: ``MeasurementTimeseriesTVPObservation`` object instance
        :return: An URL to the current object instance
        """
        # ToDo: verify it works without feature_type specified
        if "request" in self.context and self.context["request"] and obj.related_sampling_feature:
            if obj.related_sampling_feature_type in FeatureTypeEnum.values():
                path_route = r'monitoringfeature-{}s-detail'.format(''.join(obj.related_sampling_feature_type.lower().split('_')))
                # else:
                #     path_route = r'monitoringfeature-detail'
                try:
                    url = reverse(viewname=path_route,
                                  # ToDo: take off the database prefix?
                                  kwargs={'pk': obj.related_sampling_feature},
                                  request=self.context["request"], )
                except Exception:
                    return None
                return url
        return None


class FeatureSerializer(ChooseFieldsSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.Feature`
    """

    id = serializers.CharField(read_only=True)
    name = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    feature_type = serializers.CharField(read_only=True)
    observed_properties = serializers.SerializerMethodField()

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        # ToDo: Figure out what this is doing and explain it better.
        kwargs.pop('fields', None)
        super().__init__(*args, **kwargs)

        self.fields["url"] = serializers.SerializerMethodField()

    def get_url(self, obj):
        """
        Get the Site url based on the current context
        :param obj: an object instance
        :return: An URL to the current object instance
        """
        # ToDo: verify it works without feature_type specified
        if "request" in self.context and self.context["request"]:
            if obj.feature_type is not None:
                path_route = r'monitoringfeature-{}s-detail'.format(''.join(obj.feature_type.lower().split('_')))
                # else:
                # path_route = r'monitoringfeature-detail'
                try:
                    url = reverse(viewname=path_route,
                                  # ToDo: take off the database prefix?
                                  kwargs={'pk': obj.id},
                                  request=self.context["request"], )
                except Exception:
                    return None
                return url
        return None

    def get_observed_properties(self, obj):
        op_list: List = []
        if not obj.observed_properties:
            return op_list
        for mapped_attribute in obj.observed_properties:
            if mapped_attribute.get_basin3d_vocab() == NO_MAPPING_TEXT:
                datasource_vocab = mapped_attribute.get_datasource_vocab()
                logger.info(f'{obj.id} has unmapped OBSERVED_PROPERTY {datasource_vocab}')
                continue
            op = mapped_attribute.get_basin3d_desc()
            op_list.append(op.basin3d_vocab)
        op_list.sort()
        return op_list


class SamplingFeatureSerializer(FeatureSerializer):
    """
    Serializes a :class:`basin3d.core.models.SamplingFeature`
    """

    related_sampling_feature_complex = serializers.ListSerializer(
        child=ReadOnlySynthesisModelField(serializer_class=RelatedSamplingFeatureSerializer))

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)
        super().__init__(*args, **kwargs)


class SpatialSamplingFeatureSerializer(SamplingFeatureSerializer):
    """
    Serializes a :class:`basin3d.core.models.SpatialSamplingFeature`
    """

    shape = serializers.CharField(read_only=True)
    coordinates = ReadOnlySynthesisModelField(serializer_class=CoordinateSerializer)

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)
        super().__init__(*args, **kwargs)


class MonitoringFeatureSerializer(SpatialSamplingFeatureSerializer):
    """
    Serializes a :class:`basin3d.core.models.MonitoringFeature`
    """
    description_reference = serializers.CharField(read_only=True)
    related_party = serializers.ListSerializer(child=ReadOnlySynthesisModelField(serializer_class=PersonSerializer))
    utc_offset = serializers.IntegerField(read_only=True)

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)
        super().__init__(*args, **kwargs)


class ObservationSerializerMixin(object):
    """
    Serializes a :class:`basin3d.core.models.Observation`
    """

    def __init__(self, *args, **kwargs):
        super(ObservationSerializerMixin, self).__init__(*args, **kwargs)

        self.fields["id"] = serializers.CharField(read_only=True)  # type: ignore
        self.fields["type"] = serializers.CharField(read_only=True)  # type: ignore
        self.fields["utc_offset"] = serializers.IntegerField(read_only=True)  # type: ignore
        self.fields["phenomenon_time"] = TimestampField(read_only=True)  # type: ignore
        self.fields["observed_property"] = MappedAttributeField(read_only=True)  # type: ignore
        self.fields["result_quality"] = serializers.ListSerializer(child=MappedAttributeField(read_only=True))  # type: ignore
        self.fields["feature_of_interest"] = ReadOnlySynthesisModelField(serializer_class=MonitoringFeatureSerializer)  # type: ignore
        self.fields["feature_of_interest_type"] = serializers.CharField(read_only=True)  # type: ignore


class ResultListTVPSerializer(serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.ResultsListTVP`
    """
    value = serializers.SerializerMethodField()
    result_quality = serializers.ListSerializer(child=MappedAttributeField(read_only=True))

    FIELDS_OPTIONAL = {'result_quality'}

    def __init__(self, *args, **kwargs):
        """
        Override `serializers.BaseSerializer.__init__()` to modify the fields outputted.

        :param args:
        :param kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)

        super().__init__(*args, **kwargs)

        field_to_remove = set()

        instance = None
        if "instance" in kwargs:
            instance = kwargs["instance"]
        elif len(args) >= 1:
            if args[0] and isinstance(args[0], (list, tuple)) and not isinstance(args[0], str):
                instance = args[0][0]
            else:
                instance = args[0]

        if instance:
            # Remove optional fields.  We don't want them crowding the json
            for field in self.FIELDS_OPTIONAL:
                if not instance.__getattribute__(field):
                    field_to_remove.update([field])

        # remove unneeded fields
        for field in field_to_remove:
            if field in self.fields:
                self.fields.pop(field)

    def get_value(self, obj):
        """
        Get the value (i.e., the timeseries data)
        :param obj: ``MeasurementTimeseriesTVPObservation`` object instance
        :return:
        """
        return obj.value


class MeasurementTimeseriesTVPObservationSerializer(ObservationSerializerMixin, serializers.Serializer):
    """
    Serializes a :class:`basin3d.core.models.MeasurementTimeseriesTVPObservation`

    """
    aggregation_duration = MappedAttributeField(read_only=True)
    time_reference_position = serializers.CharField(read_only=True)
    sampling_medium = MappedAttributeField(read_only=True)
    statistic = MappedAttributeField(read_only=True)
    result = ReadOnlySynthesisModelField(serializer_class=ResultListTVPSerializer)
    unit_of_measurement = serializers.CharField(read_only=True)
    datasource = serializers.SerializerMethodField()

    FIELDS_OPTIONAL = {'aggregation_duration', 'time_reference_position', 'statistic', 'sampling_medium'}

    def __init__(self, *args, **kwargs):
        """
        Override `serializers.BaseSerializer.__init__()` to modify the fields outputted. Remove id if it doesn't exist

        :param args:
        :param kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        kwargs.pop('fields', None)

        super(MeasurementTimeseriesTVPObservationSerializer, self).__init__(*args, **kwargs)

        field_to_remove = set()

        instance = None
        if "instance" in kwargs:
            instance = kwargs["instance"]
        elif len(args) >= 1:
            if args[0] and isinstance(args[0], (list, tuple)) and not isinstance(args[0], str):
                instance = args[0][0]
            else:
                instance = args[0]

        if instance:
            # Remove optional fields.  We don't want them crowding the json
            if not instance.id:
                field_to_remove.update(["id", "url"])
            for field in self.FIELDS_OPTIONAL:
                if not instance.__getattribute__(field):
                    field_to_remove.update([field])

        # remove unneeded fields
        for field in field_to_remove:
            if field in self.fields:
                self.fields.pop(field)

    def get_url(self, obj):
        """
        Get the  url based on the current context
        :param obj: ``MeasurementTimeseriesTVPObservation`` object instance
        :return: An URL to the current object instance
        """
        if obj.id and "request" in self.context and self.context["request"]:
            return reverse(viewname='measurementtvptimeseries-detail', kwargs={'pk': obj.id}, request=self.context["request"], )

    def get_datasource(self, obj):
        """
        Return the url for the data sources associated with the current observation
        :param obj:
        :return:
        """
        url_kwargs = {'id_prefix': obj.datasource.id_prefix, }
        if "request" in self.context and self.context["request"]:
            return "{}".format(reverse('datasource-detail', kwargs=url_kwargs, request=self.context["request"], ))
        else:
            return obj.datasource.name



REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    'DEFAULT_PERMISSION_CLASSES': [
        # 'rest_framework.permissions.DjangoModelPermissionsOrAnonReadOnly'
    ],
    'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.AcceptHeaderVersioning',
    'DEFAULT_FILTER_BACKENDS': ('django_filters.rest_framework.DjangoFilterBackend',),
    'DEFAULT_VERSION': '1.0',
    'ALLOWED_VERSIONS': ['1.0'],
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
    ),
}

BASIN3D = {

}



"""
========
REST API
========
.. currentmodule:: django_basin3d

:platform: Unix, Mac
:synopsis: All BASIN-3D REST API calls are read-only (GET). The browsable API may be accessed at the root URL of the application.
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top


Synthesis API
*************
Synthesis is the process of converting multiple heterogeneous data sources into a single uniform format.
This section describes BASIN-3D synthesis REST API.

Data Sources
------------
BASIN-3D Data sources definitions.  All data sources defined are available for synthesis in the
subsequent APIs.

| `/datasources --` Returns a list of data sources
| `/datasources/:id_prefix --` Get a single data source

**Attributes:**
    - *name:* Unique name for the data source
    - *id_prefix:* A unique iset of characters to prefix ids for the data source
    - *location:* Location of the data source
    - *observed_property:* BASIN-3D Observed Property mappings for the data source

**URLs**
  + url -- URL with details for the data source
  + check -- Validation URL for the data source connection


Observed Property
---------------------------
BASIN-3D vocabulary for observed properties. A observed property defines what is
being measured. Data source vocabularies are mapped to these BASIN-3D vocabularies.

| `/observedproperty --` Returns a list of observed properties
| `/observedproperty/:basin3d_vocab --` Get a single observed property

**Attributes:**
    - *basin3d_vocab:* Unique observed property vocabulary
    - *full_name:* Descriptive name for the observed property
    - *categories:* Categories of which the variable is a member, listed in hierarchical order
    - *units:* Units of the observed property

**URLs**
  + url -- URL with details for the observed property


Attribute Mapping
-------------------
The Attribute Mappings registered for the Data Source plugins

| `/attributemapping --` Returns a list of all attribute mappings
| `/attributemapping/:id --` Get a single attribute mapping

**Attributes:**
    - *attr_type:* Attribute Type; e.g., STATISTIC, RESULT_QUALITY, OBSERVED_PROPERTY; separate compound mappings with ':'
    - *basin3d_vocab:* The BASIN-3D vocabulary; separate compound mappings with ':'
    - *basin3d_desc:* The BASIN-3D vocabulary descriptions; observed property objects or enum
    - *datasource_vocab:* The datasource vocabulary
    - *datasource_desc:* The datasource vocabulary description
    - *datasource:* The datasource of the mapping

**URLs**
  + url -- URL with details for the attribute mapping


Monitoring Features
-------------------
A feature on which an observation is made. Features organized into spatial
hierarchies are described via the related_sampling_feature complex

| `/monitoringfeature --` Returns a list of monitoring features types
| `/monitoringfeature/:featuretype --` Returns a list of monitoring features of the specified feature type
| `/monitoringfeature/:featuretype/:id --` Get a single monitoring feature

**Synthesis Response**
This endpoint returns the following synthesis response object.

```json
{ "query": {}, "data": [] }
```

**Data Attributes**
    Attribute for each data element from the synthesis response is as follows:

    - *id:* Unique feature identifier
    - *name:* Feature name
    - *description:* Description of the feature
    - *feature_type:* Type of feature, supported feature types: REGION, SUBREGION, BASIN, SUBBASIN, WATERSHED, SUBWATERSHED, SITE, PLOT, HORIZONTAL PATH, VERTICAL PATH, POINT
    - *observed_property_variables:* list of observed variables made at the feature. Observed property variables are configured via the plugins.
    - *related_sampling_feature_complex:* List of related sampling_features. PARENT features are currently supported.
    - *shape:* Shape of the feature: POINT, CURVE, SURFACE, SOLID
    - *coordinates:* Location of feature in absolute and/or representative datum
    - *description_reference:* Additional information about the feature
    - *related_party:* List of people or organizations responsible for the feature
    - *utc_offset:* Coordinate Universal Time offset in hours

**URLs**
  + url -- URL with details for the feature


MeasurementTimeseriesTVPObservation
-----------------------------------
MeasurementTimeseriesTVPObservation: Series of measurement (numerical) observations
in TVP (time value pair) format grouped by time (i.e., a timeseries).
Attributes specified at the group level apply to all observations.

| `/measurementtvptimeseries/?filters --` Get a single measurement timeseries TVP observation:

**Synthesis Response**
This endpoint returns the following synthesis response object.

```json
{ "query": {}, "data": [] }
```

**Data Attributes**
    Attribute for each data element from the synthesis response is as follows:

    * *id:* Observation identifier (optional)
    * *type:* MEASUREMENT_TVP_TIMESERIES
    * *observed_property:* BASIN-3D vocabulary for the observation's observed property
    * *sampling_medium:* Sampling medium of the observed property (SOLID_PHASE, WATER, GAS, OTHER)
    * *phenomenon_time:* datetime of the observation, for a timeseries the start and end times can be provided
    * *utc_offset:* Coordinate Universal Time offset in hours (offset in hours), e.g., +9
    * *feature_of_interest:* feature on which the observation is being made
    * *feature_of_interest_type:* feature type of the feature of interest
    * *result:* values = the observed values of the observed property being assessed, and result_qualit (opt) = their result_quality,
    * *time_reference_position:* position of timestamp in aggregated_duration (START, MIDDLE, END)
    * *aggregation_duration:* time period represented by observation (YEAR, MONTH, DAY, HOUR, MINUTE, SECOND)
    * *unit_of_measurement:* units in which the observation is reported
    * *statistic:* statistical property of the observation result (MEAN, MIN, MAX, TOTAL)
    * *result_quality:* quality assessments of the results (VALIDATED, UNVALIDATED, SUSPECTED, REJECTED, ESTIMATED)

**URLs**
  + url -- URL with details for the feature
  + datasource -- URL of the datasource

**Filters**
    - *monitoring_feature (required):* comma separated list of monitoring_features ids
    - *observed_property (required):* comma separated list of observed property basin3d vocabularies
    - *start_date (required):* date YYYY-MM-DD
    - *end_date (optional):* date YYYY-MM-DD
    - *aggregation_duration (default: DAY):* enum (YEAR|MONTH|DAY|HOUR|MINUTE|SECOND|NONE)
    - *statistic (optional):* comma separated list of statistic enum(s) (MEAN|MIN|MAX|INSTANTANEOUS)
    - *result_quality (optional):* comma separated list of result quality enum(s) enum (VALIDATED|UNVALIDATED|SUSPECTED|REJECTED|ESTIMATED)
    - *sampling_medium (optional):* comma separated list of sampling medium enum(s) (SOLID_PHASE|WATER|GAS|OTHER)
    - *datasource (optional):* a single data source id prefix (e.g ?datasource=`datasource.id_prefix`)

"""
from importlib.metadata import version, PackageNotFoundError
import logging

logger = logging.getLogger(__name__)

try:
    __version__ = version("django_basin3d")
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = ['get_url']

# application loads this AppConfig subclass by default
# when django_basin3d is added to INSTALLED_APPS
default_app_config = 'django_basin3d.apps.Basin3DConfig'


def __insert_basin3d_defaults():
    """
    Insert BASIN-3D default settings :class:`django_basin3d.settings`
    """

    from django.conf import global_settings, settings
    from django_basin3d import settings as basin3d_settings

    # Add the values from the application.settings module
    for key in dir(basin3d_settings):
        if key.isupper():

            # Add the defaults to the global setting
            setattr(global_settings, key, getattr(basin3d_settings, key))

            # only add default if they have not been set already
            # We don't want to override local setting
            if not hasattr(settings, key):
                setattr(settings, key, getattr(basin3d_settings, key))
            elif key in ['BASIN3D', 'REST_FRAMEWORK']:
                basin3d = getattr(basin3d_settings, key)
                local_basin3d = getattr(settings, key)
                for key in basin3d.keys():
                    if key not in local_basin3d:
                        local_basin3d[key] = basin3d[key]


__insert_basin3d_defaults()


def get_url(url, params=None, headers=None, verify=False):
    """
    Send a GET request to the specified URL

    :param url:
    :param params: request parameters
    :param headers: request headers
    :param verify: verify SSL connection
    :return: Response
    """
    import requests
    response = requests.get(url, params=params, verify=verify, headers=headers)
    logger.debug("url:{}".format(response.url))
    return response


def post_url(url, params=None, headers=None, verify=False):
    """
    Send a POST request to the specified URL

    :param url:
    :param params: request parameters
    :param headers: request headers
    :param verify: verify SSL connection
    :return: Response
    """
    import requests
    response = requests.post(url, params=params, verify=verify, headers=headers)
    logger.debug("url:{}".format(response.url))
    return response


__all__ = ['synthesis']



# Generated by Django 4.2.1 on 2023-06-13 15:45

from django.db import migrations, models
import django.db.models.deletion
import django_basin3d.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DataSource',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=20, unique=True)),
                ('id_prefix', models.CharField(max_length=5, unique=True)),
                ('location', models.TextField(blank=True)),
                ('plugin_module', models.TextField(blank=True)),
                ('plugin_class', models.TextField(blank=True)),
            ],
            options={
                'ordering': ['id_prefix'],
            },
        ),
        migrations.CreateModel(
            name='ObservedProperty',
            fields=[
                ('basin3d_vocab', models.CharField(max_length=50, primary_key=True, serialize=False, unique=True)),
                ('full_name', models.CharField(max_length=255)),
                ('categories', django_basin3d.models.StringListField(blank=True, null=True)),
                ('units', models.CharField(max_length=50)),
            ],
            options={
                'ordering': ('basin3d_vocab',),
            },
        ),
        migrations.CreateModel(
            name='AttributeMapping',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attr_type', models.CharField(max_length=50)),
                ('basin3d_vocab', models.CharField(max_length=50)),
                ('basin3d_desc', models.JSONField()),
                ('datasource_vocab', models.CharField(max_length=50)),
                ('datasource_desc', models.TextField(blank=True, null=True)),
                ('datasource', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='django_basin3d.datasource')),
            ],
            options={
                'ordering': ('datasource', 'attr_type', 'basin3d_vocab'),
                'unique_together': {('datasource', 'attr_type', 'datasource_vocab')},
            },
        ),
    ]



"""
`django_basin3d.views`
**********************

.. currentmodule:: django_basin3d.views

:platform: Unix, Mac
:synopsis: BASIN-3D Views
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top
"""
import logging
from collections import OrderedDict
from typing import Dict

from basin3d.core.schema.enum import FeatureTypeEnum
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import NoReverseMatch, reverse

from django_basin3d.models import DataSource

logger = logging.getLogger(__name__)


@api_view(['GET'])
def broker_api_root(request, format=None):
    """
    """

    # auto generated views from viewsets
    views = [('datasource', 'datasource-list'),
             ('attributemapping', 'attributemapping-list'),
             ('observedproperty', 'observedproperty-list'),
             ]
    root_dict = OrderedDict()
    # Iterate over the possible views. If they are enabled add them to the
    # root api.
    for k, v in views:
        try:
            root_dict[k] = reverse(v, request=request, format=format)
        except NoReverseMatch:
            # If there is no match just don't show it
            logger.error("NoReversMatch for {}".format(k))

    root_dict['measurementtvptimeseries'] = \
        '{}://{}/measurement_tvp_timeseries/'.format(request.scheme, request.get_host())

    root_dict['monitoringfeature'] = \
        '{}://{}/monitoringfeature/'.format(request.scheme, request.get_host())
    return Response(root_dict)


@api_view(['GET'])
def monitoring_features_lists(request, format=format):
    """
    Generate list of URLs to views for monitoring features based on availability in datasource
    """
    monitoring_features_list: Dict = {}
    supported_feature_types = FeatureTypeEnum.values()
    for datasource in DataSource.objects.all():
        viewset_models = []
        plugin = datasource.get_plugin()  # Get the plugin model

        plugin_views = plugin.get_plugin_access()
        for model_name in plugin_views.keys():
            viewset_models.append(model_name.__name__)

        datasource_feature_types = plugin.get_feature_types()
        unsupported_feature_types = []
        for feature_type in datasource_feature_types:
            if feature_type in supported_feature_types:
                ft = ''.join(feature_type.lower().split('_'))
                if ft not in monitoring_features_list.keys():
                    monitoring_features_list['{}s'.format(ft)] = \
                        '{}://{}/monitoringfeature/{}s/'.format(
                        request.scheme, request.get_host(), ft)
            elif feature_type not in unsupported_feature_types:
                unsupported_feature_types.append(feature_type)

        if len(unsupported_feature_types) > 0:
            logger.warning("{} are not supported FeatureTypes in {}.".format(", ".join(unsupported_feature_types),
                                                                             datasource.name))

    return Response(monitoring_features_list)
